import asyncio
import atexit
import concurrent.futures
import enum
import logging
import os
import platform
import shlex
import subprocess
import threading
from dataclasses import dataclass
from pathlib import Path
from typing import IO, Callable, Dict, Iterator, List, Optional, Set, Tuple, Union, cast

import psutil
import uqbar.io
import uqbar.objects

from .enums import BootStatus
from .exceptions import ServerCannotBoot
from .typing import FutureLike

logger = logging.getLogger(__name__)

DEFAULT_IP_ADDRESS = "127.0.0.1"
DEFAULT_PORT = 57110
ENVAR_SERVER_EXECUTABLE = "SUPRIYA_SERVER_EXECUTABLE"


@dataclass(frozen=True)
class Options:
    """
    SuperCollider server options configuration.
    """

    ### CLASS VARIABLES ###

    audio_bus_channel_count: int = 1024
    block_size: int = 64
    buffer_count: int = 1024
    control_bus_channel_count: int = 16384
    executable: Optional[str] = None
    hardware_buffer_size: Optional[int] = None
    initial_node_id: int = 1000
    input_bus_channel_count: int = 8
    input_device: Optional[str] = None
    input_stream_mask: str = ""
    ip_address: str = DEFAULT_IP_ADDRESS
    load_synthdefs: bool = True
    maximum_logins: int = 1
    maximum_node_count: int = 1024
    maximum_synthdef_count: int = 1024
    memory_locking: bool = False
    memory_size: int = 8192
    output_bus_channel_count: int = 8
    output_device: Optional[str] = None
    output_stream_mask: str = ""
    password: Optional[str] = None
    port: int = DEFAULT_PORT
    protocol: str = "udp"
    random_number_generator_count: int = 64
    remote_control_volume: bool = False
    realtime: bool = True
    restricted_path: Optional[str] = None
    sample_rate: Optional[int] = None
    threads: Optional[int] = None
    ugen_plugins_path: Optional[str] = None
    verbosity: int = 0
    wire_buffer_count: int = 64
    zero_configuration: bool = False

    ### INITIALIZER ###

    def __post_init__(self):
        if self.input_bus_channel_count is None:
            object.__setattr__(self, "input_bus_channel_count", 8)
        if self.output_bus_channel_count is None:
            object.__setattr__(self, "output_bus_channel_count", 8)
        if self.input_bus_channel_count < 0:
            raise ValueError(self.input_bus_channel_count)
        if self.output_bus_channel_count < 0:
            raise ValueError(self.output_bus_channel_count)
        if self.audio_bus_channel_count < (
            self.input_bus_channel_count + self.output_bus_channel_count
        ):
            raise ValueError("Insufficient audio buses")

    ### CLASS VARIABLES ###

    def __repr__(self):
        return uqbar.objects.get_repr(self, multiline=True, suppress_defaults=False)

    def __iter__(self):
        return (arg for arg in self.serialize())

    ### PUBLIC METHODS ###

    def get_audio_bus_ids(self, client_id: int) -> Tuple[int, int]:
        audio_buses_per_client = (
            self.private_audio_bus_channel_count // self.maximum_logins
        )
        minimum = self.first_private_bus_id + (client_id * audio_buses_per_client)
        maximum = self.first_private_bus_id + ((client_id + 1) * audio_buses_per_client)
        return minimum, maximum

    def get_buffer_ids(self, client_id: int) -> Tuple[int, int]:
        buffers_per_client = self.buffer_count // self.maximum_logins
        minimum = client_id * buffers_per_client
        maximum = (client_id + 1) * buffers_per_client
        return minimum, maximum

    def get_control_bus_ids(self, client_id: int) -> Tuple[int, int]:
        control_buses_per_client = self.control_bus_channel_count // self.maximum_logins
        minimum = client_id * control_buses_per_client
        maximum = (client_id + 1) * control_buses_per_client
        return minimum, maximum

    def get_sync_ids(self, client_id: int) -> Tuple[int, int]:
        return client_id << 26, (client_id + 1) << 26

    def serialize(self) -> List[str]:
        result = [str(find(self.executable))]
        pairs: Dict[str, Optional[Union[List[str], str]]] = {}
        if self.realtime:
            if self.protocol == "tcp":
                pairs["-t"] = str(self.port)
            else:
                pairs["-u"] = str(self.port)
            if self.input_device == self.output_device:
                if self.input_device:
                    pairs["-H"] = str(self.input_device)
            else:
                pairs["-H"] = [
                    str(self.input_device or ""),
                    str(self.output_device or ""),
                ]
            if self.maximum_logins != 64:
                pairs["-l"] = str(self.maximum_logins)
            if self.password:
                pairs["-p"] = str(self.password)
            if self.sample_rate is not None:
                pairs["-S"] = str(self.sample_rate)
            if not self.zero_configuration:
                pairs["-R"] = "0"
        if self.audio_bus_channel_count != 1024:
            pairs["-a"] = str(self.audio_bus_channel_count)
        if self.control_bus_channel_count != 16384:
            pairs["-c"] = str(self.control_bus_channel_count)
        if self.input_bus_channel_count != 8:
            pairs["-i"] = str(self.input_bus_channel_count)
        if self.output_bus_channel_count != 8:
            pairs["-o"] = str(self.output_bus_channel_count)
        if self.buffer_count != 1024:
            pairs["-b"] = str(self.buffer_count)
        if self.maximum_node_count != 1024:
            pairs["-n"] = str(self.maximum_node_count)
        if self.maximum_synthdef_count != 1024:
            pairs["-d"] = str(self.maximum_synthdef_count)
        if self.block_size != 64:
            pairs["-z"] = str(self.block_size)
        if self.hardware_buffer_size is not None:
            pairs["-Z"] = str(self.hardware_buffer_size)
        if self.memory_size != 8192:
            pairs["-m"] = str(self.memory_size)
        if self.random_number_generator_count != 64:
            pairs["-r"] = str(self.random_number_generator_count)
        if self.wire_buffer_count != 64:
            pairs["-w"] = str(self.wire_buffer_count)
        if not self.load_synthdefs:
            pairs["-D"] = "0"
        if self.input_stream_mask:
            pairs["-I"] = str(self.input_stream_mask)
        if self.output_stream_mask:
            pairs["-O"] = str(self.output_stream_mask)
        if 0 < self.verbosity:
            pairs["-v"] = str(self.verbosity)
        if self.restricted_path is not None:
            pairs["-P"] = str(self.restricted_path)
        if self.memory_locking:
            pairs["-L"] = None
        if self.ugen_plugins_path:
            pairs["-U"] = str(self.ugen_plugins_path)
        if self.threads and find(self.executable).stem == "supernova":
            pairs["-t"] = str(self.threads)
        for key, value in sorted(pairs.items()):
            result.append(key)
            if isinstance(value, str):
                result.append(value)
            elif isinstance(value, list):
                result.extend(value)
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def first_private_bus_id(self):
        return self.output_bus_channel_count + self.input_bus_channel_count

    @property
    def private_audio_bus_channel_count(self):
        return (
            self.audio_bus_channel_count
            - self.input_bus_channel_count
            - self.output_bus_channel_count
        )


def find(scsynth_path=None):
    """
    Find the ``scsynth`` executable.

    The following paths, if defined, will be searched (prioritised as ordered):

    1. The absolute path ``scsynth_path``
    2. The environment variable ``SUPRIYA_SERVER_EXECUTABLE`` (pointing to the `scsynth`
       binary)
    3. The user's ``PATH``
    4. Common installation directories of the SuperCollider application.

    Returns a path to the ``scsynth`` executable. Raises ``RuntimeError`` if no path is
    found.
    """
    path = Path(scsynth_path or os.environ.get(ENVAR_SERVER_EXECUTABLE) or "scsynth")
    if path.is_absolute() and uqbar.io.find_executable(str(path)):
        return path
    path_candidates = uqbar.io.find_executable(path.name)
    if path_candidates:
        return Path(path_candidates[0])
    paths = []
    executable = scsynth_path or "scsynth"
    if Path(executable).stem == "supernova":
        executable = "supernova"
    system = platform.system()
    if system == "Linux":
        paths.extend(
            [Path("/usr/bin/" + executable), Path("/usr/local/bin/" + executable)]
        )
    elif system == "Darwin":
        paths.append(
            Path("/Applications/SuperCollider.app/Contents/Resources/" + executable)
        )
    elif system == "Windows":
        paths.extend(
            Path(r"C:\Program Files").glob(r"SuperCollider*\\" + executable + ".exe")
        )
    for path in paths:
        if path.exists():
            return path
    raise RuntimeError("Failed to locate executable")


def kill():
    for process in psutil.process_iter():
        if process.name() in ("scsynth", "supernova", "scsynth.exe", "supernova.exe"):
            logger.info(f"killing {process!r}")
            process.kill()


class LineStatus(enum.IntEnum):
    CONTINUE = 0
    READY = 1
    ERROR = 2


class Capture:

    def __init__(self, process_protocol: "ProcessProtocol") -> None:
        self.process_protocol = process_protocol
        self.lines: List[str] = []

    def __enter__(self) -> "Capture":
        self.process_protocol.captures.add(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.process_protocol.captures.remove(self)

    def __iter__(self) -> Iterator[str]:
        return iter(self.lines)

    def __len__(self) -> int:
        return len(self.lines)


class ProcessProtocol:
    def __init__(
        self,
        *,
        name: Optional[str] = None,
        on_boot_callback: Optional[Callable] = None,
        on_panic_callback: Optional[Callable] = None,
        on_quit_callback: Optional[Callable] = None,
    ) -> None:
        self.buffer_ = ""
        self.captures: Set[Capture] = set()
        self.error_text = ""
        self.name = name
        self.on_boot_callback = on_boot_callback
        self.on_panic_callback = on_panic_callback
        self.on_quit_callback = on_quit_callback
        self.status = BootStatus.OFFLINE
        self.options = Options()

    def _boot(self, options: Options) -> bool:
        self.options = options
        logger.info(
            f"[{self.options.ip_address}:{self.options.port}/{self.name or hex(id(self))}] "
            "booting ..."
        )
        if self.status != BootStatus.OFFLINE:
            logger.info(
                f"[{self.options.ip_address}:{self.options.port}/{self.name or hex(id(self))}] "
                "... already booted!"
            )
            return False
        self.status = BootStatus.BOOTING
        self.error_text = ""
        self.buffer_ = ""
        logger.info(
            f"[{self.options.ip_address}:{self.options.port}/{self.name or hex(id(self))}] "
            "command: {}".format(shlex.join(options))
        )
        return True

    def _handle_data_received(
        self,
        *,
        boot_future: FutureLike[bool],
        text: str,
    ) -> Tuple[bool, bool]:
        resolved = False
        errored = False
        if "\n" in text:
            text, _, self.buffer_ = text.rpartition("\n")
            for line in text.splitlines():
                for capture in self.captures:
                    capture.lines.append(line)
                line_status = self._parse_line(line)
                if line_status == LineStatus.READY:
                    boot_future.set_result(True)
                    self.status = BootStatus.ONLINE
                    resolved = True
                    logger.info(
                        f"[{self.options.ip_address}:{self.options.port}/{self.name or hex(id(self))}] "
                        "... booted!"
                    )
                elif line_status == LineStatus.ERROR:
                    if not boot_future.done():
                        boot_future.set_result(False)
                        self.status = BootStatus.OFFLINE
                        self.error_text = line
                        resolved = True
                        errored = True
                    logger.info("... failed to boot!")
        else:
            self.buffer_ = text
        return resolved, errored

    def _parse_line(self, line: str) -> LineStatus:
        logger.info(
            f"[{self.options.ip_address}:{self.options.port}/{self.name or hex(id(self))}] "
            f"received: {line}"
        )
        if line.startswith(("SuperCollider 3 server ready", "Supernova ready")):
            return LineStatus.READY
        elif line.startswith(("Exception", "ERROR", "*** ERROR")):
            return LineStatus.ERROR
        return LineStatus.CONTINUE

    def _quit(self) -> bool:
        logger.info(
            f"[{self.options.ip_address}:{self.options.port}/{self.name or hex(id(self))}] "
            "quitting ..."
        )
        if self.status != BootStatus.ONLINE:
            logger.info(
                f"[{self.options.ip_address}:{self.options.port}/{self.name or hex(id(self))}] "
                "... already quit!"
            )
            return False
        self.status = BootStatus.QUITTING
        return True

    def capture(self) -> Capture:
        return Capture(self)


class SyncProcessProtocol(ProcessProtocol):
    def __init__(
        self,
        *,
        name: Optional[str] = None,
        on_boot_callback: Optional[Callable] = None,
        on_panic_callback: Optional[Callable] = None,
        on_quit_callback: Optional[Callable] = None,
    ) -> None:
        super().__init__(
            name=name,
            on_boot_callback=on_boot_callback,
            on_panic_callback=on_panic_callback,
            on_quit_callback=on_quit_callback,
        )
        atexit.register(self.quit)
        self.boot_future: concurrent.futures.Future[bool] = concurrent.futures.Future()
        self.exit_future: concurrent.futures.Future[int] = concurrent.futures.Future()

    def _run_process_thread(self, options: Options) -> None:
        self.process = subprocess.Popen(
            list(options),
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
            start_new_session=True,
        )
        read_thread = threading.Thread(
            args=(),
            daemon=True,
            target=self._run_read_thread,
        )
        read_thread.start()
        self.process.wait()
        was_quitting = self.status == BootStatus.QUITTING
        self.status = BootStatus.OFFLINE
        self.exit_future.set_result(self.process.returncode)
        if not self.boot_future.done():
            self.boot_future.set_result(False)
        if was_quitting and self.on_quit_callback:
            self.on_quit_callback()
        elif not was_quitting and self.on_panic_callback:
            self.on_panic_callback()

    def _run_read_thread(self) -> None:
        while self.status in (BootStatus.BOOTING, BootStatus.ONLINE):
            if not (text := cast(IO[bytes], self.process.stdout).readline().decode()):
                continue
            _, _ = self._handle_data_received(boot_future=self.boot_future, text=text)

    def _shutdown(self) -> None:
        self.process.terminate()
        self.thread.join()
        self.status = BootStatus.OFFLINE

    def boot(self, options: Options) -> None:
        if not self._boot(options):
            return
        self.boot_future = concurrent.futures.Future()
        self.exit_future = concurrent.futures.Future()
        self.thread = threading.Thread(
            args=(options,),
            daemon=True,
            target=self._run_process_thread,
        )
        self.thread.start()
        if not (self.boot_future.result()):
            self._shutdown()
            raise ServerCannotBoot(self.error_text)
        if self.on_boot_callback:
            self.on_boot_callback()

    def quit(self) -> None:
        if not self._quit():
            return
        self._shutdown()
        self.exit_future.result()
        logger.info(
            f"[{self.options.ip_address}:{self.options.port}/{self.name or hex(id(self))}] "
            "... quit!"
        )


class AsyncProcessProtocol(asyncio.SubprocessProtocol, ProcessProtocol):
    ### INITIALIZER ###

    def __init__(
        self,
        *,
        name: Optional[str] = None,
        on_boot_callback: Optional[Callable] = None,
        on_panic_callback: Optional[Callable] = None,
        on_quit_callback: Optional[Callable] = None,
    ) -> None:
        ProcessProtocol.__init__(
            self,
            name=name,
            on_boot_callback=on_boot_callback,
            on_panic_callback=on_panic_callback,
            on_quit_callback=on_quit_callback,
        )
        asyncio.SubprocessProtocol.__init__(self)
        self.boot_future: asyncio.Future[bool] = asyncio.Future()
        self.exit_future: asyncio.Future[bool] = asyncio.Future()

    ### PUBLIC METHODS ###

    async def boot(self, options: Options) -> None:
        if not self._boot(options):
            return
        loop = asyncio.get_running_loop()
        self.boot_future = loop.create_future()
        self.exit_future = loop.create_future()
        _, _ = await loop.subprocess_exec(
            lambda: self, *options, stdin=None, stderr=None
        )
        if not (await self.boot_future):
            await self.exit_future
            raise ServerCannotBoot(self.error_text)
        if self.on_boot_callback:
            self.on_boot_callback()

    def connection_made(self, transport) -> None:
        logger.info(
            f"[{self.options.ip_address}:{self.options.port}/{self.name or hex(id(self))}] "
            "connection made!"
        )
        self.transport = transport

    def pipe_connection_lost(self, fd, exc) -> None:
        logger.info(
            f"[{self.options.ip_address}:{self.options.port}/{self.name or hex(id(self))}] "
            "pipe connection lost!"
        )

    def pipe_data_received(self, fd, data) -> None:
        # *nix and OSX return full lines,
        # but Windows will return partial lines
        # which obligates us to reconstruct them.
        text = self.buffer_ + data.decode().replace("\r\n", "\n")
        _, _ = self._handle_data_received(boot_future=self.boot_future, text=text)

    def process_exited(self) -> None:
        return_code = self.transport.get_returncode()
        logger.info(
            f"[{self.options.ip_address}:{self.options.port}/{self.name or hex(id(self))}] "
            f"process exited with {return_code}."
        )
        was_quitting = self.status == BootStatus.QUITTING
        try:
            self.exit_future.set_result(return_code)
            self.status = BootStatus.OFFLINE
            if not self.boot_future.done():
                self.boot_future.set_result(False)
        except asyncio.exceptions.InvalidStateError:
            pass
        if was_quitting and self.on_quit_callback:
            self.on_quit_callback()
        elif not was_quitting and self.on_panic_callback:
            self.on_panic_callback()

    async def quit(self) -> None:
        if not self._quit():
            return
        self.transport.close()
        await self.exit_future
        logger.info(
            f"[{self.options.ip_address}:{self.options.port}/{self.name or hex(id(self))}] "
            "... quit!"
        )


class AsyncNonrealtimeProcessProtocol(asyncio.SubprocessProtocol, ProcessProtocol):
    def __init__(self) -> None:
        ProcessProtocol.__init__(self)
        asyncio.SubprocessProtocol.__init__(self)
        self.boot_future: asyncio.Future[bool] = asyncio.Future()
        self.exit_future: asyncio.Future[bool] = asyncio.Future()

    async def run(self, command: List[str], render_directory_path: Path) -> None:
        logger.info(f"running: {shlex.join(command)}")
        loop = asyncio.get_running_loop()
        self.boot_future = loop.create_future()
        self.exit_future = loop.create_future()
        _, _ = await loop.subprocess_exec(
            lambda: self,
            *command,
            stdin=None,
            stderr=None,
            start_new_session=True,
            cwd=render_directory_path,
        )
        await self.exit_future

    def connection_made(self, transport) -> None:
        logger.info("connection made!")
        self.transport = transport

    def pipe_connection_lost(self, fd, exc) -> None:
        logger.info("pipe connection lost!")

    def pipe_data_received(self, fd, data) -> None:
        # *nix and OSX return full lines,
        # but Windows will return partial lines
        # which obligates us to reconstruct them.
        text = self.buffer_ + data.decode().replace("\r\n", "\n")
        _, _ = self._handle_data_received(boot_future=self.boot_future, text=text)

    def process_exited(self) -> None:
        return_code = self.transport.get_returncode()
        logger.info(f"process exited with {return_code}.")
        self.exit_future.set_result(return_code)
        try:
            if not self.boot_future.done():
                self.boot_future.set_result(False)
        except asyncio.exceptions.InvalidStateError:
            pass



from typing import Any, Dict, Tuple

from ..enums import CalculationRate, DoneAction
from .core import UGen, param, ugen


@ugen(ar=True, kr=True, is_multichannel=True)
class BufRd(UGen):
    """
    A buffer-reading oscillator.

    ::

        >>> buffer_id = 23
        >>> phase = supriya.ugens.Phasor.ar(
        ...     rate=supriya.ugens.BufRateScale.kr(buffer_id=buffer_id),
        ...     start=0,
        ...     stop=supriya.ugens.BufFrames.kr(buffer_id=buffer_id),
        ... )
        >>> buf_rd = supriya.ugens.BufRd.ar(
        ...     buffer_id=buffer_id,
        ...     channel_count=2,
        ...     interpolation=2,
        ...     loop=1,
        ...     phase=phase,
        ... )
        >>> buf_rd
        <BufRd.ar()>
    """

    buffer_id = param()
    phase = param(0.0)
    loop = param(1)
    interpolation = param(2)


@ugen(ar=True, kr=True, has_done_flag=True)
class BufWr(UGen):
    """
    A buffer-writing oscillator.

    ::

        >>> buffer_id = 23
        >>> phase = supriya.ugens.Phasor.ar(
        ...     rate=supriya.ugens.BufRateScale.kr(buffer_id=buffer_id),
        ...     start=0,
        ...     stop=supriya.ugens.BufFrames.kr(buffer_id=buffer_id),
        ... )
        >>> source = supriya.ugens.In.ar(bus=0, channel_count=2)
        >>> buf_wr = supriya.ugens.BufWr.ar(
        ...     buffer_id=buffer_id,
        ...     loop=1,
        ...     phase=phase,
        ...     source=source,
        ... )
        >>> buf_wr
        <BufWr.ar()[0]>
    """

    buffer_id = param()
    phase = param(0.0)
    loop = param(1.0)
    source = param(unexpanded=True)


@ugen(ir=True, is_width_first=True)
class ClearBuf(UGen):
    """
    ::

        >>> clear_buf = supriya.ugens.ClearBuf.ir(
        ...     buffer_id=23,
        ... )
        >>> clear_buf
        <ClearBuf.ir()[0]>
    """

    buffer_id = param()


@ugen(ir=True, is_width_first=True)
class LocalBuf(UGen):
    """
    A synth-local buffer.

    ::

        >>> from supriya.ugens import FFT, IFFT, LocalBuf, Out, PinkNoise, SynthDefBuilder

    ::

        >>> local_buf = LocalBuf.ir(
        ...     channel_count=1,
        ...     frame_count=1,
        ... )
        >>> local_buf
        <LocalBuf.ir()[0]>

    LocalBuf creates a ``MaxLocalBufs`` UGen implicitly during SynthDef compilation:

    ::

        >>> with SynthDefBuilder() as builder:
        ...     local_buf = LocalBuf.ir(frame_count=2048)
        ...     source = PinkNoise.ar()
        ...     pv_chain = FFT.kr(
        ...         buffer_id=local_buf,
        ...         source=source,
        ...     )
        ...     ifft = IFFT.ar(pv_chain=pv_chain)
        ...     out = Out.ar(bus=0, source=ifft)
        ...
        >>> synthdef = builder.build()
        >>> for ugen in synthdef.ugens:
        ...     ugen
        ...
        <MaxLocalBufs.ir()>
        <LocalBuf.ir()>
        <PinkNoise.ar()>
        <FFT.kr()>
        <IFFT.ar()>
        <Out.ar()>
    """

    channel_count = param(1)
    frame_count = param(1)

    def _postprocess_kwargs(
        self, *, calculation_rate: CalculationRate, **kwargs
    ) -> Tuple[CalculationRate, Dict[str, Any]]:
        return CalculationRate.SCALAR, kwargs


@ugen(ir=True)
class MaxLocalBufs(UGen):
    """
    Sets the maximum number of local buffers in a synth.

    Used internally by LocalBuf.

    ::

        >>> max_local_bufs = supriya.ugens.MaxLocalBufs.ir(maximum=1)
        >>> max_local_bufs
        <MaxLocalBufs.ir()[0]>
    """

    maximum = param(0)


@ugen(ar=True, kr=True, is_multichannel=True)
class PlayBuf(UGen):
    """
    A sample playback oscillator.

    ::

        >>> buffer_id = 23
        >>> play_buf = supriya.ugens.PlayBuf.ar(
        ...     buffer_id=buffer_id,
        ...     channel_count=2,
        ...     done_action=0,
        ...     loop=0,
        ...     rate=1,
        ...     start_position=0,
        ...     trigger=1,
        ... )
        >>> play_buf
        <PlayBuf.ar()>
    """

    buffer_id = param()
    rate = param(1)
    trigger = param(1)
    start_position = param(0)
    loop = param(0)
    done_action = param(0)


@ugen(ar=True, kr=True, has_done_flag=True)
class RecordBuf(UGen):
    """
    Records or overdubs into a buffer.

    ::

        >>> buffer_id = 23
        >>> source = supriya.ugens.In.ar(bus=0, channel_count=2)
        >>> record_buf = supriya.ugens.RecordBuf.ar(
        ...     buffer_id=buffer_id,
        ...     done_action=0,
        ...     loop=1,
        ...     offset=0,
        ...     preexisting_level=0,
        ...     record_level=1,
        ...     run=1,
        ...     source=source,
        ...     trigger=1,
        ... )
        >>> record_buf
        <RecordBuf.ar()[0]>
    """

    buffer_id = param()
    offset = param(0.0)
    record_level = param(1.0)
    preexisting_level = param(0.0)
    run = param(1.0)
    loop = param(1.0)
    trigger = param(1.0)
    done_action = param(DoneAction(0))
    source = param(unexpanded=True)



import pytest

import supriya

pytest_plugins = ["sphinx.testing.fixtures"]


@pytest.fixture
def server():
    server = supriya.Server()
    server.set_latency(0.0)
    server.boot()
    server.add_synthdefs(supriya.default)
    server.sync()
    yield server
    server.quit()


@pytest.fixture(scope="module")
def persistent_server():
    server = supriya.Server()
    server.set_latency(0.0)
    server.boot()
    yield server
    server.quit()



"""
Utility functions.

These will be migrated out into a base package at some point.
"""

from .intervals import Interval, IntervalTree, Moment
from .iterables import (
    expand,
    flatten,
    group_by_count,
    iterate_nwise,
    repeat_to_length,
    zip_cycled,
)

__all__ = [
    "Interval",
    "IntervalTree",
    "Moment",
    "expand",
    "flatten",
    "group_by_count",
    "iterate_nwise",
    "repeat_to_length",
    "zip_cycled",
]



import logging

import pytest

from supriya.contexts.errors import ContextError, InvalidCalculationRate
from supriya.contexts.nonrealtime import Score
from supriya.osc import OscBundle, OscMessage


@pytest.fixture(autouse=True)
def use_caplog(caplog):
    caplog.set_level(logging.INFO)


@pytest.fixture
def context():
    return Score()


def test_add_bus(context):
    # no moment
    with pytest.raises(ContextError):
        context.add_bus("AUDIO")
    with context.at(0):
        # invalid calculation rate
        with pytest.raises(InvalidCalculationRate):
            context.add_bus("SCALAR")
        # ok
        audio_bus = context.add_bus("AUDIO")
        control_bus = context.add_bus("CONTROL")
    assert audio_bus.context is context
    assert audio_bus.id_ == 16
    assert control_bus.context is context
    assert control_bus.id_ == 0
    assert list(context.iterate_osc_bundles()) == []


def test_add_bus_group(context):
    # no moment
    with pytest.raises(ContextError):
        context.add_bus_group("AUDIO", 0)
    with context.at(0):
        # invalid calculation rate
        with pytest.raises(InvalidCalculationRate):
            context.add_bus_group("SCALAR")
        # count less than 1
        with pytest.raises(ValueError):
            context.add_bus_group("AUDIO", 0)
        # ok
        audio_bus_group = context.add_bus_group("AUDIO", 8)
        control_bus_group = context.add_bus_group("CONTROL", 4)
    assert len(audio_bus_group) == 8
    assert all(audio_bus.context is context for audio_bus in audio_bus_group)
    assert [audio_bus.id_ for audio_bus in audio_bus_group] == [
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
    ]
    assert len(control_bus_group) == 4
    assert all(control_bus.context is context for control_bus in control_bus_group)
    assert [control_bus.id_ for control_bus in control_bus_group] == [0, 1, 2, 3]
    assert list(context.iterate_osc_bundles()) == []


def test_fill_buses(context):
    with context.at(0):
        audio_bus = context.add_bus("AUDIO")
        control_bus_a = context.add_bus("CONTROL")
        control_bus_b = context.add_bus("CONTROL")
        control_bus_c = context.add_bus("CONTROL")
        with pytest.raises(InvalidCalculationRate):
            audio_bus.fill(2, 0.75)
        control_bus_a.fill(3, 0.5)
    with context.at(1.23):
        control_bus_b.fill(4, 0.25)
        control_bus_c.fill(5, 0.125)
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(contents=(OscMessage("/c_fill", 0, 3, 0.5),), timestamp=0.0),
        OscBundle(
            contents=(OscMessage("/c_fill", 1, 4, 0.25, 2, 5, 0.125),), timestamp=1.23
        ),
    ]


def test_free_bus(context):
    with context.at(0):
        audio_bus = context.add_bus("AUDIO")
        control_bus = context.add_bus("CONTROL")
        audio_bus.free()
        control_bus.free()
    assert list(context.iterate_osc_bundles()) == []
    with context.at(1.23):
        new_audio_bus = context.add_bus("AUDIO")
        new_control_bus = context.add_bus("CONTROL")
    # NRT contexts don't re-use IDs
    assert new_audio_bus.id_ == audio_bus.id_ + 1
    assert new_control_bus.id_ == control_bus.id_ + 1


def test_set_bus(context):
    with context.at(0):
        audio_bus = context.add_bus("AUDIO")
        control_bus_a = context.add_bus("CONTROL")
        control_bus_b = context.add_bus("CONTROL")
        control_bus_c = context.add_bus("CONTROL")
        with pytest.raises(InvalidCalculationRate):
            audio_bus.set(0.75)
        control_bus_a.set(0.5)
    with context.at(1.23):
        control_bus_b.set(0.25)
        control_bus_c.set(0.125)
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(contents=(OscMessage("/c_set", 0, 0.5),), timestamp=0.0),
        OscBundle(contents=(OscMessage("/c_set", 1, 0.25, 2, 0.125),), timestamp=1.23),
    ]


@pytest.mark.asyncio
async def test_set_bus_range(context):
    with context.at(0):
        audio_bus_group = context.add_bus_group("AUDIO", count=4)
        control_bus_group = context.add_bus_group("CONTROL", count=4)
    with context.at(1.23):
        with pytest.raises(InvalidCalculationRate):
            audio_bus_group[0].set_range((0.1, 0.2, 0.3, 0.4))
        control_bus_group[0].set_range((0.1, 0.2, 0.3, 0.4))
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(OscMessage("/c_setn", 0, 4, 0.1, 0.2, 0.3, 0.4),), timestamp=1.23
        )
    ]



"""
Classes for modeling responses from :term:`scsynth`.
"""

import dataclasses
import re
from collections import deque
from typing import (
    Deque,
    Dict,
    Generator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)

from ..enums import NodeAction
from ..osc import OscMessage


@dataclasses.dataclass
class Response:
    @classmethod
    def from_osc(cls, osc_message: OscMessage) -> "Response":
        mapping: Dict[str, Type[Response]] = {
            "/b_info": BufferInfo,
            "/b_set": GetBufferInfo,
            "/b_setn": GetBufferRangeInfo,
            "/c_set": GetControlBusInfo,
            "/c_setn": GetControlBusRangeInfo,
            "/d_removed": SynthDefRemovedInfo,
            "/done": DoneInfo,
            "/fail": FailInfo,
            "/g_queryTree.reply": QueryTreeInfo,
            "/n_end": NodeInfo,
            "/n_go": NodeInfo,
            "/n_info": NodeInfo,
            "/n_move": NodeInfo,
            "/n_off": NodeInfo,
            "/n_on": NodeInfo,
            "/n_set": GetNodeControlInfo,
            "/n_setn": GetNodeControlRangeInfo,
            "/status.reply": StatusInfo,
            "/synced": SyncedInfo,
            "/tr": TriggerInfo,
            "/version.reply": VersionInfo,
        }
        return mapping[str(osc_message.address)].from_osc(osc_message)


@dataclasses.dataclass
class BufferInfo(Response):
    """
    A ``/b_info`` response.
    """

    @dataclasses.dataclass
    class Item:
        buffer_id: int
        frame_count: int
        channel_count: int
        sample_rate: float

    items: Sequence[Item]

    @classmethod
    def from_osc(cls, osc_message: OscMessage) -> "Response":
        return cls(
            items=[
                cls.Item(*osc_message.contents[i : i + 4])
                for i in range(0, len(osc_message.contents), 4)
            ]
        )


@dataclasses.dataclass
class DoneInfo(Response):
    """
    A ``/done`` response.
    """

    command_name: str
    other: Sequence[Union[float, str]]

    @classmethod
    def from_osc(cls, osc_message: OscMessage) -> "Response":
        command_name, *other = osc_message.contents
        return cls(command_name=command_name, other=other)


@dataclasses.dataclass
class FailInfo(Response):
    """
    A ``/fail`` response.
    """

    command_name: str
    error: str
    other: Sequence[Union[float, str]]

    @classmethod
    def from_osc(cls, osc_message: OscMessage) -> "Response":
        return cls(
            command_name=osc_message.contents[0],
            error=osc_message.contents[1],
            other=osc_message.contents[2:],
        )


@dataclasses.dataclass
class GetBufferInfo(Response):
    """
    A ``/b_set`` response.
    """

    buffer_id: int
    items: Sequence[Tuple[int, float]]

    @classmethod
    def from_osc(cls, osc_message: OscMessage) -> "Response":
        buffer_id = osc_message.contents[0]
        items: List[Tuple[int, float]] = []
        for i in range(1, len(osc_message.contents), 2):
            index, value = osc_message.contents[i : i + 2]
            items.append((int(index), float(value)))
        return cls(buffer_id=buffer_id, items=items)


@dataclasses.dataclass
class GetBufferRangeInfo(Response):
    """
    A ``/b_setn`` response.
    """

    buffer_id: int
    items: Sequence[Tuple[int, Sequence[float]]]

    @classmethod
    def from_osc(cls, osc_message: OscMessage) -> "Response":
        buffer_id = osc_message.contents[0]
        items: List[Tuple[int, Sequence[float]]] = []
        current_index = 1
        while current_index < len(osc_message.contents):
            index, count = osc_message.contents[current_index : current_index + 2]
            current_index += 2
            values = osc_message.contents[current_index : current_index + count]
            items.append((index, tuple(values)))
            current_index += count
        return cls(buffer_id=buffer_id, items=items)


@dataclasses.dataclass
class GetControlBusInfo(Response):
    """
    A ``/c_set`` response.
    """

    items: Sequence[Tuple[int, float]]

    @classmethod
    def from_osc(cls, osc_message: OscMessage) -> "Response":
        items: List[Tuple[int, float]] = []
        for i in range(0, len(osc_message.contents), 2):
            index, value = osc_message.contents[i : i + 2]
            items.append((int(index), float(value)))
        return cls(items=items)


@dataclasses.dataclass
class GetControlBusRangeInfo(Response):
    """
    A ``/c_setn`` response.
    """

    items: Sequence[Tuple[int, Sequence[float]]]

    @classmethod
    def from_osc(cls, osc_message: OscMessage) -> "Response":
        items: List[Tuple[int, Sequence[float]]] = []
        current_index = 0
        while current_index < len(osc_message.contents):
            index, count = osc_message.contents[current_index : current_index + 2]
            current_index += 2
            values = osc_message.contents[current_index : current_index + count]
            items.append((index, tuple(values)))
            current_index += count
        return cls(items=items)


@dataclasses.dataclass
class GetNodeControlInfo(Response):
    node_id: int
    items: Sequence[Tuple[Union[int, str], float]]

    @classmethod
    def from_osc(cls, osc_message: OscMessage) -> "Response":
        node_id, *rest = osc_message.contents
        items: List[Tuple[Union[int, str], float]] = []
        for i in range(1, len(osc_message.contents), 2):
            name_or_index, value = osc_message.contents[i : i + 2]
            items.append(
                (
                    (
                        name_or_index
                        if isinstance(name_or_index, str)
                        else int(name_or_index)
                    ),
                    float(value),
                )
            )
        return cls(node_id=node_id, items=items)


@dataclasses.dataclass
class GetNodeControlRangeInfo(Response):
    node_id: int
    items: Sequence[Tuple[Union[int, str], Sequence[float]]]

    @classmethod
    def from_osc(cls, osc_message: OscMessage) -> "Response":
        node_id, *rest = osc_message.contents
        items: List[Tuple[Union[int, str], Sequence[float]]] = []
        current_index = 1
        while current_index < len(osc_message.contents):
            name_or_index, count = osc_message.contents[
                current_index : current_index + 2
            ]
            current_index += 2
            values = osc_message.contents[current_index : current_index + count]
            items.append(
                (
                    (
                        name_or_index
                        if isinstance(name_or_index, str)
                        else int(name_or_index)
                    ),
                    tuple(values),
                )
            )
            current_index += count
        return cls(node_id=node_id, items=items)


@dataclasses.dataclass
class NodeInfo(Response):
    action: NodeAction
    node_id: int
    parent_id: int
    previous_id: int
    next_id: int
    is_group: bool
    head_id: Optional[int] = None
    tail_id: Optional[int] = None

    @classmethod
    def from_osc(cls, osc_message: OscMessage) -> "Response":
        action = osc_message.address
        (
            node_id,
            parent_id,
            previous_id,
            next_id,
            is_group,
            *rest,
        ) = osc_message.contents
        if is_group:
            head_id: Optional[int] = int(rest[0])
            tail_id: Optional[int] = int(rest[1])
        else:
            head_id, tail_id = None, None
        return cls(
            action=NodeAction.from_expr(action),
            node_id=int(node_id),
            parent_id=int(parent_id),
            previous_id=int(previous_id),
            next_id=int(next_id),
            is_group=bool(is_group),
            head_id=head_id,
            tail_id=tail_id,
        )


@dataclasses.dataclass
class QueryTreeInfo(Response):
    """
    A ``/g_queryTree.reply`` response.
    """

    @dataclasses.dataclass
    class Item:
        node_id: int
        child_count: int
        synthdef_name: Optional[str] = None
        controls: Optional[Dict[Union[int, str], Union[float, str]]] = None

    node_id: int
    child_count: int
    items: Sequence[Item]

    @classmethod
    def from_osc(cls, osc_message: OscMessage) -> "Response":
        flag = osc_message.contents[0]
        node_id = int(osc_message.contents[1])
        child_count = int(osc_message.contents[2])
        items: List["QueryTreeInfo.Item"] = []
        index = 3
        while index < len(osc_message.contents):
            child_id = int(osc_message.contents[index])
            child_child_count = int(osc_message.contents[index + 1])
            synthdef_name: Optional[str] = None
            controls: Optional[Dict[Union[int, str], Union[float, str]]] = None
            index += 2
            if child_child_count < 0:
                synthdef_name = osc_message.contents[index]
                index += 1
                if flag:
                    controls = {}
                    control_count = osc_message.contents[index]
                    index += 1
                    for i in range(control_count):
                        name_or_index: Union[int, str] = osc_message.contents[index]
                        value: Union[float, str] = osc_message.contents[index + 1]
                        controls[name_or_index] = value
                        index += 2
            items.append(
                cls.Item(
                    node_id=child_id,
                    child_count=child_child_count,
                    controls=controls,
                    synthdef_name=synthdef_name,
                )
            )
        return cls(node_id=node_id, child_count=child_count, items=items)


@dataclasses.dataclass
class QueryTreeControl:
    name_or_index: Union[int, str]
    value: Union[float, str]

    def __str__(self):
        value = self.value
        try:
            value = round(value, 6)
        except Exception:
            pass
        return f"{self.name_or_index}: {value!s}"


@dataclasses.dataclass
class QueryTreeNode:
    node_id: int
    annotation: Optional[str] = None


@dataclasses.dataclass
class QueryTreeSynth(QueryTreeNode):
    node_id: int
    synthdef_name: Optional[str] = None
    controls: List[QueryTreeControl] = dataclasses.field(default_factory=list)
    annotation: Optional[str] = None

    def __format__(self, format_spec):
        return "\n".join(
            self._get_str_format_pieces(unindexed=format_spec == "unindexed")
        )

    def __str__(self):
        return "\n".join(self._get_str_format_pieces())

    def _get_str_format_pieces(self, unindexed=False):
        result = []
        node_id = self.node_id
        if unindexed:
            node_id = "..."
        string = f"{node_id} {self.synthdef_name}"
        if self.annotation:
            string += f" ({self.annotation})"
        result.append(string)
        if self.controls:
            result.append("    " + ", ".join(str(control) for control in self.controls))
        return result


@dataclasses.dataclass
class QueryTreeGroup(QueryTreeNode):
    node_id: int
    children: List[QueryTreeNode] = dataclasses.field(default_factory=list)
    annotation: Optional[str] = None

    ### SPECIAL METHODS ###

    def __format__(self, format_spec):
        return "NODE TREE " + "\n".join(
            self._get_str_format_pieces(unindexed=format_spec == "unindexed")
        )

    def __str__(self):
        return "NODE TREE " + "\n".join(self._get_str_format_pieces())

    ### PRIVATE METHODS ###

    def _get_str_format_pieces(self, unindexed=False):
        result = []
        node_id = self.node_id
        if unindexed:
            node_id = "..."
        string = f"{node_id} group"
        if self.annotation:
            string += f" ({self.annotation})"
        result.append(string)
        for child in self.children:
            for line in child._get_str_format_pieces(unindexed=unindexed):
                result.append("    {}".format(line))
        return result

    ### PUBLIC METHODS ###

    def annotate(self, annotations: Dict[int, str]) -> "QueryTreeGroup":
        root = self
        if root.node_id in annotations:
            root = dataclasses.replace(root, annotation=annotations[root.node_id])
        for group in root.walk():
            for i, child in enumerate(group.children):
                if child.node_id in annotations:
                    group.children[i] = dataclasses.replace(
                        child, annotation=annotations[child.node_id]
                    )
        return root

    @classmethod
    def from_query_tree_info(cls, response: QueryTreeInfo) -> "QueryTreeGroup":
        def recurse(
            item: QueryTreeInfo.Item, items: Deque[QueryTreeInfo.Item]
        ) -> Union[QueryTreeGroup, QueryTreeSynth]:
            if item.child_count < 0:
                return QueryTreeSynth(
                    node_id=item.node_id,
                    synthdef_name=item.synthdef_name,
                    controls=[
                        QueryTreeControl(name_or_index=name_or_index, value=value)
                        for name_or_index, value in (item.controls or {}).items()
                    ],
                )
            children: List[QueryTreeNode] = []
            for _ in range(item.child_count):
                children.append(recurse(items.popleft(), items))
            return QueryTreeGroup(node_id=item.node_id, children=children)

        return cast(
            QueryTreeGroup,
            recurse(
                QueryTreeInfo.Item(
                    node_id=response.node_id, child_count=response.child_count
                ),
                deque(response.items),
            ),
        )

    @classmethod
    def from_string(cls, string) -> "QueryTreeGroup":
        node_pattern = re.compile(r"^\s*(\d+) (\S+)$")
        control_pattern = re.compile(r"\w+: \S+")
        lines = string.splitlines()
        if not lines[0].startswith("NODE TREE"):
            raise ValueError
        stack: List[QueryTreeGroup] = [
            QueryTreeGroup(node_id=int(lines.pop(0).rpartition(" ")[-1]))
        ]
        for line in lines:
            indent = line.count("   ")
            if match := (node_pattern.match(line)):
                while len(stack) > indent:
                    stack.pop()
                node_id = int(match.groups()[0])
                if (name := match.groups()[1]) == "group":
                    stack[-1].children.append(group := QueryTreeGroup(node_id=node_id))
                    stack.append(group)
                else:
                    stack[-1].children.append(
                        synth := QueryTreeSynth(node_id=node_id, synthdef_name=name)
                    )
            else:
                for pair in control_pattern.findall(line):
                    name_string, _, value_string = pair.partition(": ")
                    try:
                        name_or_index: Union[int, str] = int(name_string)
                    except ValueError:
                        name_or_index = name_string
                    try:
                        value: Union[float, str] = float(value_string)
                    except ValueError:
                        value = value_string
                    synth.controls.append(
                        QueryTreeControl(name_or_index=name_or_index, value=value)
                    )
        return stack[0]

    def walk(self) -> Generator["QueryTreeGroup", None, None]:
        yield self
        for child in self.children:
            if isinstance(child, QueryTreeGroup):
                yield from child.walk()


@dataclasses.dataclass
class StatusInfo(Response):
    """
    A ``/status.reply`` response.
    """

    actual_sample_rate: float
    average_cpu_usage: float
    group_count: int
    peak_cpu_usage: float
    synth_count: int
    synthdef_count: int
    target_sample_rate: float
    ugen_count: int

    @classmethod
    def from_osc(cls, osc_message: OscMessage) -> "Response":
        (
            _,
            ugen_count,
            synth_count,
            group_count,
            synthdef_count,
            average_cpu_usage,
            peak_cpu_usage,
            target_sample_rate,
            actual_sample_rate,
        ) = osc_message.contents
        return cls(
            actual_sample_rate=actual_sample_rate,
            average_cpu_usage=average_cpu_usage,
            group_count=group_count,
            peak_cpu_usage=peak_cpu_usage,
            synth_count=synth_count,
            synthdef_count=synthdef_count,
            target_sample_rate=target_sample_rate,
            ugen_count=ugen_count,
        )


@dataclasses.dataclass
class SyncedInfo(Response):
    """
    A ``/synced`` response.
    """

    sync_id: int

    @classmethod
    def from_osc(cls, osc_message: OscMessage) -> "Response":
        return cls(*osc_message.contents)


@dataclasses.dataclass
class SynthDefRemovedInfo(Response):
    """
    A ``/d_removed`` response.
    """

    name: str

    @classmethod
    def from_osc(cls, osc_message: OscMessage) -> "Response":
        return cls(*osc_message.contents)


@dataclasses.dataclass
class TriggerInfo(Response):
    """
    A ``/tr`` response.
    """

    node_id: int
    trigger_id: int
    value: float

    @classmethod
    def from_osc(cls, osc_message: OscMessage) -> "Response":
        return cls(*osc_message.contents)


@dataclasses.dataclass
class VersionInfo(Response):
    """
    A ``/version.reply`` response.
    """

    program_name: str
    major: int
    minor: int
    patch: str
    branch: str
    commit: str

    @classmethod
    def from_osc(cls, osc_message: OscMessage) -> "Response":
        return cls(*osc_message.contents)



from .core import UGen, param, ugen


@ugen(ar=True, is_multichannel=True)
class GrainBuf(UGen):
    """
    ::

        >>> grain_buf = supriya.ugens.GrainBuf.ar(
        ...     channel_count=2,
        ...     duration=1,
        ...     envelope_buffer_id=-1,
        ...     interpolate=2,
        ...     maximum_overlap=512,
        ...     pan=0,
        ...     position=0,
        ...     rate=1,
        ...     buffer_id=0,
        ...     trigger=0,
        ... )
        >>> grain_buf
        <GrainBuf.ar()>
    """

    trigger = param(0)
    duration = param(1)
    buffer_id = param()
    rate = param(1)
    position = param(0)
    interpolate = param(2)
    pan = param(0)
    envelope_buffer_id = param(-1)
    maximum_overlap = param(512)


@ugen(ar=True, is_multichannel=True)
class GrainIn(UGen):
    """
    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> grain_in = supriya.ugens.GrainIn.ar(
        ...     channel_count=2,
        ...     duration=1,
        ...     envelope_buffer_id=-1,
        ...     maximum_overlap=512,
        ...     position=0,
        ...     source=source,
        ...     trigger=0,
        ... )
        >>> grain_in
        <GrainIn.ar()>
    """

    trigger = param(0)
    duration = param(1)
    source = param()
    position = param(0)
    envelope_buffer_id = param(-1)
    maximum_overlap = param(512)


@ugen(ar=True)
class PitchShift(UGen):
    """
    A pitch shift unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.PitchShift.ar(
        ...     source=source,
        ... )
        <PitchShift.ar()[0]>
    """

    source = param()
    window_size = param(0.2)
    pitch_ratio = param(1.0)
    pitch_dispersion = param(0.0)
    time_dispersion = param(0.0)


@ugen(ar=True, is_multichannel=True)
class Warp1(UGen):
    """
    ::

        >>> warp_1 = supriya.ugens.Warp1.ar(
        ...     buffer_id=0,
        ...     channel_count=1,
        ...     envelope_buffer_id=-1,
        ...     frequency_scaling=1,
        ...     interpolation=1,
        ...     overlaps=8,
        ...     pointer=0,
        ...     window_rand_ratio=0,
        ...     window_size=0.2,
        ... )
        >>> warp_1
        <Warp1.ar()[0]>
    """

    buffer_id = param(0)
    pointer = param(0)
    frequency_scaling = param(1)
    window_size = param(0.2)
    envelope_buffer_id = param(-1)
    overlaps = param(8)
    window_rand_ratio = param(0)
    interpolation = param(1)



"""
Tools for interacting with non-realtime execution contexts.
"""

import hashlib
import logging
import platform
import shutil
import struct
from contextlib import ExitStack
from os import PathLike
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, Iterator, List, Optional, SupportsInt, Tuple, Type, Union

from uqbar.objects import new

from ..assets.synthdefs import system_synthdefs
from ..enums import BootStatus, CalculationRate, HeaderFormat, SampleFormat
from ..osc import OscBundle
from ..scsynth import AsyncNonrealtimeProcessProtocol, Options
from ..typing import HeaderFormatLike, SampleFormatLike, SupportsOsc
from ..ugens import SynthDef
from .core import Context, ContextError, ContextObject, Node
from .requests import DoNothing, RequestBundle, Requestable

logger = logging.getLogger(__name__)


class Score(Context):
    """
    A non-realtime execution context.

    :param options: The context's options.
    :param kwargs: Keyword arguments for options.
    """

    ### INITIALIZER ###

    def __init__(self, options: Optional[Options] = None, **kwargs):
        super().__init__(options=options, **kwargs)
        self._boot_status = BootStatus.ONLINE
        self._requests: Dict[float, List[Requestable]] = {}

    ### CLASS METHODS ###

    async def __render__(
        self,
        *,
        output_file_path: Optional[PathLike] = None,
        render_directory_path: Optional[PathLike] = None,
        **kwargs,
    ) -> Tuple[Optional[Path], int]:
        return await self.render(
            output_file_path,
            render_directory_path=render_directory_path,
            **kwargs,
        )

    ### PRIVATE METHODS ###

    def _free_id(
        self,
        type_: Type[ContextObject],
        id_: int,
        calculation_rate: Optional[CalculationRate] = None,
    ) -> None:
        pass

    def _resolve_node(self, node: Union[Node, SupportsInt, None]) -> int:
        if node is None:
            return 0
        return int(node)

    def _validate_can_request(self) -> None:
        if self._get_moment() is None:
            raise ContextError

    def _validate_moment_timestamp(self, seconds: Optional[float]) -> None:
        if seconds is None or seconds < 0:
            raise ContextError

    ### PUBLIC METHODS ###

    async def render(
        self,
        output_file_path: Optional[PathLike] = None,
        *,
        duration: Optional[float] = None,
        header_format: HeaderFormatLike = HeaderFormat.AIFF,
        input_file_path: Optional[PathLike] = None,
        options: Optional[Options] = None,
        render_directory_path: Optional[PathLike] = None,
        sample_format: SampleFormatLike = SampleFormat.INT24,
        sample_rate: float = 44100,
        suppress_output: bool = False,
        **kwargs,
    ) -> Tuple[Optional[Path], int]:
        """
        Render the score.

        :param duration: Optional duration to render the score until.
        :param header_format: The :term:`header format` to render with.
        :param input_file_path: The input file to render with.
        :param options: The context's options.
        :param render_directory_path: The directory to render the output in. This
            affords using relative paths and (therefore) stable hashes when rendering
            multiple scores that interrelate.
        :param sample_format: The :term:`sample format` to render with.
        :param sample_rate: The sample rate to render at.
        :param suppress_output: Flag for writing the output soundfile to ``/dev/null``
            (or equivalent).
        :param kwargs: Keyword arguments for options.

        :return: A pair of the output path (if output exists) and the process exit code.
            If no output file path was provided, one will be generated based on the hash
            of the score's datagram, its input file (if provided) and any flags to
            ``scsynth`` that affect rendering.
        """
        from .. import output_path

        # validate inputs
        header_format_ = HeaderFormat.from_expr(header_format).name.lower()
        sample_format_ = SampleFormat.from_expr(sample_format).name.lower()
        # build initial command
        command = new(options or self._options, **kwargs, realtime=False).serialize()
        # build datagram
        datagram_pieces: List[bytes] = []
        for datagram_piece in self.iterate_datagrams(until=duration):
            datagram_pieces.append(struct.pack(">i", len(datagram_piece)))
            datagram_pieces.append(datagram_piece)
        datagram = b"".join(datagram_pieces)
        # setup render directory
        exit_stack = ExitStack()
        with exit_stack:
            # setup render directory
            if render_directory_path:
                render_directory_path_ = Path(render_directory_path).resolve()
            else:
                render_directory_path_ = Path(
                    exit_stack.enter_context(TemporaryDirectory())
                )
            logger.info(f"Render directory: {render_directory_path_}")
            # calculate input path relative to render directory
            input_file_path_ = "_"  # underscore, not dash
            if input_file_path:
                input_file_path_ = str(Path(input_file_path).resolve())
            # build file name
            hasher = hashlib.sha256()
            hasher.update(datagram)
            hasher.update((" ".join(command[1:])).encode())
            hasher.update(input_file_path_.encode())
            hasher.update(str(sample_rate).encode())
            digest = hasher.hexdigest()
            # build render file path and output file path
            if suppress_output:
                render_file_name = (
                    "NUL" if platform.system() == "Windows" else "/dev/null"
                )
                output_file_path_ = None
            else:
                render_file_name = f"score-{digest}.{header_format_}"
                output_file_path_ = Path(
                    output_file_path or (output_path / render_file_name)
                )
            # write .osc file
            osc_file_name = f"score-{digest}.osc"
            (render_directory_path_ / osc_file_name).write_bytes(datagram)
            # build nonrealtime command
            command.extend(
                [
                    "-N",
                    osc_file_name,
                    input_file_path_,
                    render_file_name,
                    str(sample_rate),
                    header_format_,
                    sample_format_,
                ]
            )
            # render the datagram
            protocol = AsyncNonrealtimeProcessProtocol()
            await protocol.run(command, render_directory_path_)
            exit_code: int = await protocol.exit_future
            assert render_directory_path_ / render_file_name
            if output_file_path_:
                shutil.copy(
                    render_directory_path_ / render_file_name, output_file_path_
                )
        return output_file_path_, exit_code

    def iterate_datagrams(self, until: Optional[float] = None) -> Iterator[bytes]:
        """
        Iterate datagrams.

        :param until: Timestamp to stop iterating at.
        """
        for osc_bundle in self.iterate_osc_bundles(until=until):
            yield osc_bundle.to_datagram(realtime=False)

    def iterate_osc_bundles(self, until: Optional[float] = None) -> Iterator[OscBundle]:
        """
        Iterate OSC bundles.

        :param until: Timestamp to stop iterating at.
        """
        for request_bundle in self.iterate_request_bundles(until=until):
            yield request_bundle.to_osc()

    def iterate_request_bundles(
        self, until: Optional[float] = None
    ) -> Iterator[RequestBundle]:
        """
        Iterate request bundles.

        :param until: Timestamp to stop iterating at.
        """
        for timestamp, requests in sorted(self._requests.items()):
            if until:
                if timestamp == until:
                    yield RequestBundle(
                        timestamp=timestamp, contents=requests + [DoNothing()]
                    )
                    return
                elif timestamp > until:
                    yield RequestBundle(timestamp=until, contents=[DoNothing()])
                    return
            if requests:
                yield RequestBundle(timestamp=timestamp, contents=requests)
        if until and until > timestamp:
            yield RequestBundle(timestamp=until, contents=[DoNothing()])

    def send(self, message: SupportsOsc) -> None:
        """
        Send a message to the execution context.

        :param message: The message to send.
        """
        if not isinstance(message, RequestBundle):
            raise ContextError
        elif message.timestamp is None:
            raise ContextError
        self._requests.setdefault(message.timestamp, []).extend(message.contents)

    def setup_system_synthdefs(self) -> None:
        """
        Load all system synthdefs.
        """
        synthdefs = []
        for name in dir(system_synthdefs):
            synthdef = getattr(system_synthdefs, name)
            if isinstance(synthdef, SynthDef):
                synthdefs.append(synthdef)
        with self.at(0):
            self.add_synthdefs(*synthdefs)



cdef extern from "server_shm.hpp" namespace "detail_server_shm":
    cdef cppclass server_shared_memory_client:
        server_shared_memory_client(unsigned int port_number) except +
        float* get_control_busses()
        void set_control_bus(int bus, float value)



import asyncio
import logging

import pytest

from supriya.clocks import AsyncOfflineClock, TimeUnit

repeat_count = 5

logger = logging.getLogger("supriya.test")


@pytest.fixture(autouse=True)
def log_everything(caplog):
    caplog.set_level(logging.INFO, logger="supriya.clocks")


async def callback(
    context,
    store,
    blow_up_at=None,
    delta=0.25,
    limit=4,
    time_unit=TimeUnit.BEATS,
    **kwargs,
):
    if context.event.invocations == blow_up_at:
        raise Exception
    store.append((context.current_moment, context.desired_moment, context.event))
    if limit is None:
        return delta, time_unit
    elif context.event.invocations < limit:
        return delta, time_unit
    return None


def check(store):
    moments = []
    for current_moment, desired_moment, event in store:
        one = [
            "{}/{}".format(*current_moment.time_signature),
            current_moment.beats_per_minute,
        ]
        two = [
            current_moment.measure,
            round(current_moment.measure_offset, 10),
            current_moment.offset,
            current_moment.seconds,
        ]
        three = [
            desired_moment.measure,
            round(desired_moment.measure_offset, 10),
            desired_moment.offset,
            desired_moment.seconds,
        ]
        moments.append((one, two, three))
    return moments


@pytest.mark.asyncio
async def test_basic():
    clock = AsyncOfflineClock()
    store = []
    clock.schedule(callback, schedule_at=0.0, args=[store])
    await clock.start()
    await asyncio.sleep(0.1)
    assert check(store) == [
        (["4/4", 120.0], [1, 0.0, 0.0, 0.0], [1, 0.0, 0.0, 0.0]),
        (["4/4", 120.0], [1, 0.25, 0.25, 0.5], [1, 0.25, 0.25, 0.5]),
        (["4/4", 120.0], [1, 0.5, 0.5, 1.0], [1, 0.5, 0.5, 1.0]),
        (["4/4", 120.0], [1, 0.75, 0.75, 1.5], [1, 0.75, 0.75, 1.5]),
        (["4/4", 120.0], [2, 0.0, 1.0, 2.0], [2, 0.0, 1.0, 2.0]),
    ]
    await clock.stop()



import pytest

from supriya.patterns import EventPattern, NoteEvent, SequencePattern
from supriya.patterns.testutils import MockUUID as M
from supriya.patterns.testutils import run_pattern_test


@pytest.mark.parametrize(
    "stop_at, input_a, input_b, expected, is_infinite",
    [
        (
            None,
            SequencePattern([1, 2, 3], None),
            SequencePattern([4, 5], None),
            [
                NoteEvent(M("A"), a=1, b=4),
                NoteEvent(M("B"), a=2, b=5),
                NoteEvent(M("C"), a=3, b=4),
                NoteEvent(M("D"), a=1, b=5),
                NoteEvent(M("E"), a=2, b=4),
                NoteEvent(M("F"), a=3, b=5),
            ],
            True,
        ),
        (
            None,
            SequencePattern([1, 2, 3], None),
            SequencePattern([4, 5], 1),
            [NoteEvent(M("A"), a=1, b=4), NoteEvent(M("B"), a=2, b=5)],
            False,
        ),
        (
            None,
            SequencePattern([1, 2, 3], None),
            SequencePattern([4, 5], 2),
            [
                NoteEvent(M("A"), a=1, b=4),
                NoteEvent(M("B"), a=2, b=5),
                NoteEvent(M("C"), a=3, b=4),
                NoteEvent(M("D"), a=1, b=5),
            ],
            False,
        ),
        (
            None,
            SequencePattern([1, 2, 3], 1),
            SequencePattern([4, 5], 1),
            [NoteEvent(M("A"), a=1, b=4), NoteEvent(M("B"), a=2, b=5)],
            False,
        ),
        (
            None,
            SequencePattern([1, 2, 3], 1),
            SequencePattern([4, 5], None),
            [
                NoteEvent(M("A"), a=1, b=4),
                NoteEvent(M("B"), a=2, b=5),
                NoteEvent(M("C"), a=3, b=4),
            ],
            False,
        ),
        (
            None,
            SequencePattern([1, 2, 3], 1),
            4,
            [
                NoteEvent(M("A"), a=1, b=4),
                NoteEvent(M("B"), a=2, b=4),
                NoteEvent(M("C"), a=3, b=4),
            ],
            False,
        ),
    ],
)
def test(stop_at, input_a, input_b, expected, is_infinite):
    pattern = EventPattern(a=input_a, b=input_b)
    run_pattern_test(pattern, expected, is_infinite, stop_at)



from ...enums import DoneAction, ParameterRate
from ...ugens import (
    LPF,
    Linen,
    Mix,
    OffsetOut,
    Pan2,
    Parameter,
    Rand,
    SynthDef,
    SynthDefBuilder,
    VarSaw,
    XLine,
)


def _build_default_synthdef() -> SynthDef:
    with SynthDefBuilder(
        amplitude=0.1,
        frequency=440,
        gate=1,
        out=Parameter(rate=ParameterRate.SCALAR, value=0),
        pan=0.5,
    ) as builder:
        low_pass = LPF.ar(
            source=Mix.new(
                VarSaw.ar(
                    frequency=builder["frequency"]
                    + (
                        0,
                        Rand.ir(minimum=-0.4, maximum=0.0),
                        Rand.ir(minimum=0.0, maximum=0.4),
                    ),
                    width=0.3,
                )
            )
            * 0.3,
            frequency=XLine.kr(
                start=Rand.ir(minimum=4000, maximum=5000),
                stop=Rand.ir(minimum=2500, maximum=3200),
            ),
        )
        linen = Linen.kr(
            attack_time=0.01,
            done_action=DoneAction.FREE_SYNTH,
            gate=builder["gate"],
            release_time=0.3,
            sustain_level=0.7,
        )
        pan = Pan2.ar(
            source=low_pass * linen * builder["amplitude"], position=builder["pan"]
        )
        OffsetOut.ar(bus=builder["out"], source=pan)
    return builder.build(name="default")


default = _build_default_synthdef()

__all__ = ["default"]



from uqbar.strings import normalize

from supriya import Server, default


def test_QueryTreeGroup_annotate() -> None:
    server = Server().boot()
    server.add_synthdefs(default)
    server.sync()
    group_a = server.add_group()
    group_b = group_a.add_group()
    group_a.add_synth(synthdef=default)
    group_b.add_synth(synthdef=default)
    group_b.add_synth(synthdef=default)
    group_b.add_group()
    tree = server.query_tree()
    assert tree is not None
    assert str(tree) == normalize(
        """
        NODE TREE 0 group
            1 group
                1000 group
                    1002 default
                        amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5, out: 0.0
                    1001 group
                        1005 group
                        1004 default
                            amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5, out: 0.0
                        1003 default
                            amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5, out: 0.0
        """
    )
    assert str(tree.annotate({})) == normalize(
        """
        NODE TREE 0 group
            1 group
                1000 group
                    1002 default
                        amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5, out: 0.0
                    1001 group
                        1005 group
                        1004 default
                            amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5, out: 0.0
                        1003 default
                            amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5, out: 0.0
        """
    )
    assert str(
        tree.annotate(
            {
                0: "root",
                1: "default group",
                1000: "mixer",
                1005: "grandchild",
                1003: "synth",
            }
        )
    ) == normalize(
        """
        NODE TREE 0 group (root)
            1 group (default group)
                1000 group (mixer)
                    1002 default
                        amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5, out: 0.0
                    1001 group
                        1005 group (grandchild)
                        1004 default
                            amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5, out: 0.0
                        1003 default (synth)
                            amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5, out: 0.0
        """
    )



# flake8: noqa
import os
import platform

import pytest

from supriya.ugens import (
    Out,
    RandID,
    RandSeed,
    SuperColliderSynthDef,
    SynthDefBuilder,
    WhiteNoise,
)


@pytest.fixture
def py_synthdef_01():
    with SynthDefBuilder(rand_id=0, seed=0) as builder:
        RandID.ir(rand_id=builder["rand_id"])
        RandSeed.ir(seed=builder["seed"], trigger=1)
        source = WhiteNoise.ar()
        Out.ar(bus=0, source=source)
    py_synthdef = builder.build("seedednoise")
    return py_synthdef


@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_rngs_01_supriya_vs_sclang(py_synthdef_01):
    sc_synthdef = SuperColliderSynthDef(
        "seedednoise",
        r"""
        arg rand_id=0, seed=0;
        RandID.ir(rand_id);
        RandSeed.ir(1, seed);
        Out.ar(0, WhiteNoise.ar());
        """,
    )
    sc_compiled_synthdef = sc_synthdef.compile()
    py_compiled_synthdef = py_synthdef_01.compile()
    assert py_compiled_synthdef == sc_compiled_synthdef


def test_rngs_01_supriya_vs_bytes(py_synthdef_01):
    # fmt: off
    test_compiled_synthdef = bytes(
        b'SCgf'
        b'\x00\x00\x00\x02'
        b'\x00\x01'
            b'\x0bseedednoise'
                b'\x00\x00\x00\x02'
                    b'?\x80\x00\x00'
                    b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x02'
                    b'\x00\x00\x00\x00'
                    b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x02'
                    b'\x07rand_id'
                    b'\x00\x00\x00\x00'
                    b'\x04seed'
                    b'\x00\x00\x00\x01'
                b'\x00\x00\x00\x05'
                    b'\x07Control'
                        b'\x01'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00'
                            b'\x01'
                            b'\x01'
                    b'\x06RandID'
                        b'\x00'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x00'
                            b'\x00'
                    b'\x08RandSeed'
                        b'\x00'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x01'
                            b'\x00'
                    b'\nWhiteNoise'
                        b'\x02'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x02'
                    b'\x03Out'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x01'
                            b'\x00\x00\x00\x03'
                            b'\x00\x00\x00\x00'
                b'\x00\x00'
    )
    # fmt: on
    py_compiled_synthdef = py_synthdef_01.compile()
    assert py_compiled_synthdef == test_compiled_synthdef



import dataclasses
import tempfile
from os import PathLike
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Awaitable,
    Callable,
    Container,
    Dict,
    List,
    Optional,
    Sequence,
    SupportsFloat,
    Tuple,
    Union,
    cast,
)

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

from ..assets.synthdefs.default import default
from ..enums import AddAction, CalculationRate
from ..io import PlayMemo
from ..typing import AddActionLike, HeaderFormatLike, SampleFormatLike, SupportsRender
from ..ugens import SynthDef
from .errors import InvalidCalculationRate, InvalidMoment
from .responses import BufferInfo, NodeInfo, QueryTreeGroup

if TYPE_CHECKING:
    import numpy

    from .core import Completion, Context
    from .realtime import AsyncServer, BaseServer, Server


@dataclasses.dataclass(frozen=True)
class ContextObject:
    r"""
    Base class for objects with context references.

    :param context: The context object's context.
    :param id\_: The context object's context ID.
    """

    context: "Context"
    id_: int

    def __float__(self) -> float:
        """
        Get the context object's ID as a float.
        """
        return float(self.id_)

    def __int__(self) -> int:
        """
        Get the context object's ID as an integer.
        """
        return self.id_

    @property
    def allocated(self) -> bool:
        """
        Get the context object's allocation status.
        """
        return self in cast("BaseServer", self.context)


@dataclasses.dataclass(frozen=True)
class Buffer(ContextObject):
    r"""
    A buffer.

    :param context: The context object's context.
    :param id\_: The context object's context ID.
    :param completion: The buffer's allocation completion.
    """

    completion: Optional["Completion"] = dataclasses.field(default=None, compare=False)

    def __enter__(self) -> "Completion":
        """
        Enter the buffer's allocation completion.
        """
        if self.completion is None:
            raise InvalidMoment
        return self.completion.__enter__()

    def __exit__(self, *args) -> None:
        """
        Exit the buffer's allocation completion.
        """
        if self.completion is None:
            raise InvalidMoment
        return self.completion.__exit__(*args)

    def __plot__(self) -> Tuple["numpy.ndarray", float]:
        import librosa

        with tempfile.TemporaryDirectory() as temp_directory:
            file_path = Path(temp_directory) / "tmp.wav"
            self.write(file_path=file_path, header_format="wav", sample_format="int32")
            cast("Server", self.context).sync()
            return librosa.load(file_path, mono=False, sr=None)

    def __render_memo__(
        self,
        output_file_path: Optional[PathLike] = None,
        render_directory_path: Optional[PathLike] = None,
        **kwargs,
    ) -> SupportsRender:
        with tempfile.TemporaryDirectory() as temp_directory:
            path = Path(temp_directory) / "tmp.wav"
            self.write(file_path=path, header_format="wav", sample_format="int32")
            cast("Server", self.context).sync()
            return PlayMemo.from_path(path)

    def close(
        self, on_completion: Optional[Callable[["Context"], None]] = None
    ) -> "Completion":
        """
        Close the buffer.

        Emit ``/b_close`` requests.

        :param on_completion: A callable with the buffer's context as the only argument.
            Permits building an "on completion" argument to this method's request
            without an active moment.
        """
        return self.context.close_buffer(self, on_completion=on_completion)

    def copy(
        self,
        *,
        target_buffer: "Buffer",
        starting_frame: int,
        target_starting_frame: int,
        frame_count: int,
    ) -> None:
        """
        Copy the buffer.

        Emit ``/b_gen <buffer.id_> copy ...`` requests.

        :param target_buffer: The buffer to copy to.
        :param starting_frame: The frame index in the this buffer to start reading from.
        :param target_starting_frame: The frame index in the target buffer to start
            writing at.
        :param frame_count: The number of frames to copy.
        """
        self.context.copy_buffer(
            frame_count=frame_count,
            source_buffer=self,
            source_starting_frame=starting_frame,
            target_buffer=target_buffer,
            target_starting_frame=target_starting_frame,
        )

    def fill(self, starting_frame: int, frame_count: int, value: float) -> None:
        """
        Fill the buffer with a single value.

        Emit ``/b_fill`` requests.

        :param starting_frame: The frame index to start filling at.
        :param frame_count: The number of frames to fill.
        :param value: The value to fill with.
        """
        self.context.fill_buffer(self, starting_frame, frame_count, value)

    def free(
        self, on_completion: Optional[Callable[["Context"], None]] = None
    ) -> "Completion":
        """
        Free the buffer.

        Emit ``/b_free`` requests.

        :param on_completion: A callable with the buffer's context as the only argument.
            Permits building an "on completion" argument to this method's request
            without an active moment.
        """
        return self.context.free_buffer(self, on_completion=on_completion)

    def generate(
        self,
        command_name: Literal["sine1", "sine2", "sine3", "cheby"],
        amplitudes: Sequence[float],
        frequencies: Optional[Sequence[float]] = None,
        phases: Optional[Sequence[float]] = None,
        as_wavetable: bool = False,
        should_clear_first: bool = False,
        should_normalize: bool = False,
    ) -> None:
        """
        Generate the buffer.

        Emit ``/b_gen`` requests.

        :param command_name: The generation command name.
        :param amplitudes: A sequence of partial amplitudes.
        :param frequencies: A sequence of partial frequencies.
        :param phases: A sequence of partial phases.
        :param as_wavetable: Flag for generating the output in wavetable format.
        :param should_clear_first: Flag for clearing the buffer before generating.
        :param should_normalize: Flag for normalizing the generated output.
        """
        self.context.generate_buffer(
            buffer=self,
            command_name=command_name,
            amplitudes=amplitudes,
            frequencies=frequencies,
            phases=phases,
            as_wavetable=as_wavetable,
            should_clear_first=should_clear_first,
            should_normalize=should_normalize,
        )

    def get(
        self, *indices: int, sync: bool = True
    ) -> Union[Awaitable[Optional[Dict[int, float]]], Optional[Dict[int, float]]]:
        """
        Get a sample.

        Emit ``/b_get`` requests.

        :param indices: The sample indices to read.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        return cast(Union["AsyncServer", "Server"], self.context).get_buffer(
            self, *indices, sync=sync
        )

    def get_range(
        self, index: int, count: int, sync: bool = True
    ) -> Union[Awaitable[Optional[Sequence[float]]], Optional[Sequence[float]]]:
        """
        Get a sample range.

        Emit ``/b_getn`` requests.

        :param index: The sample index to start reading at.
        :param count: The number of samples to read.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        return cast(Union["AsyncServer", "Server"], self.context).get_buffer_range(
            self, index, count, sync=sync
        )

    def normalize(self, new_maximum: float = 1.0, as_wavetable: bool = False) -> None:
        """
        Normalize the buffer.

        Emit ``/b_gen <buffer.id_> (w)?normalize`` requests depending on parameters.

        :param new_maximum: The new maximum to normalize to.
        :param as_wavetable: Flag for treating the buffer contents as a wavetable.
        """
        self.context.normalize_buffer(
            self, new_maximum=new_maximum, as_wavetable=as_wavetable
        )

    def query(
        self, sync: bool = True
    ) -> Union[Awaitable[Optional[BufferInfo]], Optional[BufferInfo]]:
        """
        Query the buffer.

        Emit ``/b_query`` requests.

        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        return cast(Union["AsyncServer", "Server"], self.context).query_buffer(
            self, sync=sync
        )

    def read(
        self,
        file_path: PathLike,
        *,
        buffer_starting_frame: Optional[int] = None,
        channel_indices: Optional[List[int]] = None,
        frame_count: Optional[int] = None,
        leave_open: bool = False,
        on_completion: Optional[Callable[["Context"], None]] = None,
        starting_frame: Optional[int] = None,
    ) -> "Completion":
        """
        Read a file into the buffer.

        Emit ``/b_read`` or ``/b_readChannel`` requests, depending on parameters.

        :param file_path: The file path to read from.
        :param channel_indices: A list of channel indices to read from when reading from
            a file.
        :param frame_count: The number of frames to read.
        :param leave_open: Flag for leaving the file open (e.g. to continue reading via
            :py:class:`~supriya.ugens.diskio.DiskIn`) or close it.
        :param starting_frame: The starting frame in the buffer to begin reading into
            at.
        :param on_completion: A callable with the buffer's context as the only argument.
            Permits building an "on completion" argument to this method's request
            without an active moment.
        """
        return self.context.read_buffer(
            self,
            file_path,
            buffer_starting_frame=buffer_starting_frame,
            channel_indices=channel_indices,
            frame_count=frame_count,
            leave_open=leave_open,
            on_completion=on_completion,
            starting_frame=starting_frame,
        )

    def set(self, index: int, value: float) -> None:
        """
        Set a sample.

        Emit ``/b_set`` requests.

        :param index: The sample index to write at.
        :param value: The value to write.
        """
        self.context.set_buffer(buffer=self, index=index, value=value)

    def set_range(self, index: int, values: Sequence[float]) -> None:
        """
        Set a sample range.

        Emit ``/b_setn`` requests.

        :param index: The sample index to start writing at.
        :param values: The values to write.
        """
        self.context.set_buffer_range(buffer=self, index=index, values=values)

    def write(
        self,
        file_path: PathLike,
        *,
        frame_count: Optional[int] = None,
        header_format: HeaderFormatLike = "aiff",
        leave_open: bool = False,
        on_completion: Optional[Callable[["Context"], None]] = None,
        sample_format: SampleFormatLike = "int24",
        starting_frame: Optional[int] = None,
    ) -> "Completion":
        """
        Write the buffer to disk.

        Emit ``/b_write`` requests.

        :param file_path: The file path to write into.
        :param frame_count: The number of frames to write.
        :param header_format: The header format to use, e.g. ``AIFF`` or ``WAVE``.
        :param leave_open: Flag for leaving the file open (e.g. to continue writing via
            :py:class:`~supriya.ugens.diskio.DiskOut`) or close it.
        :param sample_format: The sample format to use, e.g. ``INT24`` or ``FLOAT``.
        :param starting_frame: The starting frame in the buffer to start writing from.
        :param on_completion: A callable with the buffer's context as the only argument.
            Permits building an "on completion" argument to this method's request
            without an active moment.
        """
        return self.context.write_buffer(
            self,
            file_path,
            frame_count=frame_count,
            header_format=header_format,
            leave_open=leave_open,
            on_completion=on_completion,
            sample_format=sample_format,
            starting_frame=starting_frame,
        )

    def zero(
        self, on_completion: Optional[Callable[["Context"], None]] = None
    ) -> "Completion":
        """
        Zero the buffer.

        Emit ``/b_zero`` requests.

        :param on_completion: A callable with the buffer's context as the only argument.
            Permits building an "on completion" argument to this method's request
            without an active moment.
        """
        return self.context.zero_buffer(self, on_completion=on_completion)


@dataclasses.dataclass(frozen=True)
class BufferGroup(ContextObject):
    r"""
    A buffer group.

    :param context: The buffer group's context.
    :param id\_: The buffer group's context ID.
    :param count: The number of child buffers.
    """

    count: int = 1
    buffers: Tuple[Buffer, ...] = dataclasses.field(
        init=False, repr=False, default_factory=tuple
    )

    def __post_init__(self):
        object.__setattr__(
            self,
            "buffers",
            tuple(
                Buffer(context=self.context, id_=i)
                for i in range(self.id_, self.id_ + self.count)
            ),
        )

    def __getitem__(self, item):
        return self.buffers[item]

    def __len__(self) -> int:
        return len(self.buffers)

    def free(self):
        """
        Free the buffer group.

        Emit ``/b_free`` requests.
        """
        self.context.free_buffer_group(self)


@dataclasses.dataclass(frozen=True)
class Bus(ContextObject):
    r"""
    A bus.

    :param context: The bus' context.
    :param id\_: The bus' context ID.
    :param calculation_rate: The bus' calculation rate.
    """

    calculation_rate: CalculationRate

    def fill(self, count: int, value: float) -> None:
        """
        Fill contiguous buses with a single value, starting with this bus.

        Emit ``/c_fill`` requests.

        :param count: The number of buses to fill.
        :param value: The value to fill with.
        """
        self.context.fill_bus_range(self, count, value)

    def free(self) -> None:
        """
        Free the bus.

        Emit no requests.
        """
        self.context.free_bus(self)

    def get(
        self, sync: bool = True
    ) -> Union[Awaitable[Optional[float]], Optional[float]]:
        """
        Get the control bus' value.

        Emit ``/c_get`` requests.

        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        return cast(Union["AsyncServer", "Server"], self.context).get_bus(
            self, sync=sync
        )

    def get_range(
        self, count: int, sync: bool = True
    ) -> Union[Awaitable[Optional[Sequence[float]]], Optional[Sequence[float]]]:
        """
        Get a range of control bus values.

        Emit ``/c_getn`` requests.

        :param count: The number of contiguous buses whose values to get.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        return cast(Union["AsyncServer", "Server"], self.context).get_bus_range(
            self, count, sync=sync
        )

    def map_symbol(self) -> str:
        """
        Get the bus' map symbol.
        """
        if self.calculation_rate is CalculationRate.AUDIO:
            return f"a{self.id_}"
        elif self.calculation_rate is CalculationRate.CONTROL:
            return f"c{self.id_}"
        raise InvalidCalculationRate

    def set(self, value: float) -> None:
        """
        Set the control bus's value.

        Emit ``/c_set`` requests.

        :param value: The value to set the control bus to.
        """
        self.context.set_bus(self, value)

    def set_range(self, values: Sequence[float]) -> None:
        """
        Set a range of control buses.

        Emit ``/c_setn`` requests.

        :param values: The values to write.
        """
        self.context.set_bus_range(self, values)


@dataclasses.dataclass(frozen=True)
class BusGroup(ContextObject):
    r"""
    A bus group.

    :param context: The bus group's context.
    :param id\_: The bus group's context ID.
    :param calculation_rate: The bus group's calculation rate.
    :param count: The number of child buses.
    """

    calculation_rate: CalculationRate
    count: int = 1
    buses: Tuple[Bus, ...] = dataclasses.field(
        init=False, repr=False, default_factory=tuple
    )

    def __post_init__(self):
        object.__setattr__(
            self,
            "buses",
            tuple(
                Bus(calculation_rate=self.calculation_rate, context=self.context, id_=i)
                for i in range(self.id_, self.id_ + self.count)
            ),
        )

    def __getitem__(self, item):
        return self.buses[item]

    def __len__(self) -> int:
        return len(self.buses)

    def free(self):
        """
        Free the bus group.

        Emit no requests.
        """
        self.context.free_bus_group(self)

    def map_symbol(self) -> str:
        """
        Get the bus group's map symbol.
        """
        if self.calculation_rate is CalculationRate.AUDIO:
            return f"a{self.id_}"
        elif self.calculation_rate is CalculationRate.CONTROL:
            return f"c{self.id_}"
        raise InvalidCalculationRate


@dataclasses.dataclass(frozen=True)
class Node(ContextObject):
    r"""
    A node.

    :param context: The node's context.
    :param id\_: The node's context ID.
    """

    def add_group(
        self,
        add_action: AddActionLike = AddAction.ADD_TO_HEAD,
        parallel: bool = False,
        permanent: bool = False,
    ) -> "Group":
        """
        Add a new group relative to this node.

        Emit ``/g_new`` or ``/p_new`` requests depending on parameters.

        :param add_action: The :term:`add action` to use when placing the new group.
        :param parallel: Flag for parallel vs non-parallel groups.
        :param permanent: Flag for using a permanent node ID.
        """
        return self.context.add_group(
            add_action=add_action,
            parallel=parallel,
            permanent=permanent,
            target_node=self,
        )

    def add_synth(
        self,
        synthdef: SynthDef,
        *,
        add_action: AddActionLike = AddAction.ADD_TO_HEAD,
        permanent: bool = False,
        **settings,
    ) -> "Synth":
        """
        Add a new synth relative to this node.

        Emit ``/s_new`` requests.

        :param synthdef: The :term:`SynthDef` to use for the new synth.
        :param add_action: The :term:`add action` to use when placing the new synth.
        :param permanent: Flag for using a permanent node ID.
        :param settings: The new synth's control settings.
        """
        return self.context.add_synth(
            synthdef=synthdef,
            add_action=add_action,
            permanent=permanent,
            target_node=self,
            **settings,
        )

    def free(self, force: bool = False) -> None:
        """
        Free the node.

        Emit ``/n_free`` for groups, for synths without a ``gate`` control, or when
        ``force`` is ``True``.

        Emit ``/n_set <node.id_> gate 0`` for synths with ``gate`` controls.

        :param force: Flag for force-freeing, without releasing.
        """
        return self.context.free_node(self, force=force)

    def map(self, **settings: Union[Bus, None]) -> None:
        """
        Map the node's controls to buses.

        Emit ``/n_map`` and ``/n_mapa`` requests.

        :param settings: A mapping of control names to buses (or to ``None`` to unmap
            the control).
        """
        self.context.map_node(self, **settings)

    def move(self, target_node: "Node", add_action: AddActionLike = None) -> None:
        """
        Move the node.

        Emit ``/n_after``, ``/n_before``, ``/g_head`` and ``/g_tail`` requests depending
        on parameters.

        :param add_action: The :term:`add action` to use when moving the node.
        :param target_node: The target node to place the node relative to.
        """
        self.context.move_node(
            node=self, add_action=add_action, target_node=target_node
        )

    def order(self, *nodes: "Node", add_action: AddActionLike = None) -> None:
        """
        Re-order nodes relative to the node.

        Emit ``/n_order`` requests.

        :param nodes: The nodes to re-order.
        :param add_action: The :term:`add action` to use when re-ordering the nodes.
        """
        self.context.order_nodes(self, *nodes, add_action=add_action)

    def pause(self) -> None:
        """
        Pause the node.

        Emit ``/n_run <node.id_> 0`` requests.
        """
        self.context.pause_node(self)

    def query(
        self, sync: bool = True
    ) -> Union[Awaitable[Optional[NodeInfo]], Optional[NodeInfo]]:
        """
        Query the node.

        Emit ``/n_query`` requests.

        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        return cast(Union["AsyncServer", "Server"], self.context).query_node(
            self, sync=sync
        )

    def set(
        self,
        *indexed_settings: Tuple[int, Union[SupportsFloat, Sequence[SupportsFloat]]],
        **settings: Union[SupportsFloat, Sequence[SupportsFloat]],
    ) -> None:
        """
        Set the node's controls.

        :param indexed_settings: A sequence of control indices to values.
        :param settings: A mapping of control names to values.
        """
        self.context.set_node(self, *indexed_settings, **settings)

    def set_range(
        self,
        *indexed_settings: Tuple[int, Sequence[SupportsFloat]],
        **settings: Sequence[SupportsFloat],
    ) -> None:
        """
        Set a range of the node's controls.

        :param indexed_settings: A sequence of control indices to values.
        :param settings: A mapping of control names to values.
        """
        self.context.set_node_range(self, *indexed_settings, **settings)

    def unpause(self) -> None:
        """
        Unpause the node.

        Emit ``/n_run <node.id_> 1`` requests.
        """
        self.context.unpause_node(self)

    @property
    def _valid_add_actions(self) -> Container[int]:
        return (
            AddAction.ADD_AFTER,
            AddAction.ADD_BEFORE,
            AddAction.ADD_TO_HEAD,
            AddAction.ADD_TO_TAIL,
            AddAction.REPLACE,
        )

    @property
    def active(self) -> bool:
        """
        Get the node's paused/unpaused status.
        """
        return cast("BaseServer", self.context)._node_active.get(self.id_, True)

    @property
    def parent(self) -> Optional["Group"]:
        """
        Get the node's parent, as currently cached on the context.
        """
        parent_id = cast("BaseServer", self.context)._node_parents.get(self.id_)
        if parent_id is None:
            return None
        elif parent_id == 0:
            return RootNode(context=self.context, id_=0)
        return Group(context=self.context, id_=parent_id)

    @property
    def parentage(self) -> Sequence["Node"]:
        """
        Get the node's parentage, as currently cached on the context.
        """
        context = cast("BaseServer", self.context)
        parentage: List["Node"] = [self]
        while (parent_id := context._node_parents.get(parentage[-1].id_)) is not None:
            if parent_id:
                parentage.append(Group(context=context, id_=parent_id))
            else:
                parentage.append(context.root_node)
        return parentage


@dataclasses.dataclass(frozen=True)
class Group(Node):
    r"""
    A group node.

    :param context: The group's context.
    :param id\_: The group's context ID.
    :param parallel: Flag for parallel vs non-parallel groups.
    """

    parallel: bool = False

    def dump_tree(
        self,
        include_controls: bool = True,
        sync: bool = True,
    ) -> Union[Awaitable[Optional[QueryTreeGroup]], Optional[QueryTreeGroup]]:
        """
        Dump the group's node tree.

        Emit ``/g_dumpTree`` requests.

        :param include_controls: Flag for including synth control values.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        return cast(Union["AsyncServer", "Server"], self.context).dump_tree(
            group=self, include_controls=include_controls, sync=sync
        )

    def free_children(self, synths_only: bool = False) -> None:
        """
        Free the group's children.

        Emit ``/g_deepFree`` or ``/g_freeAll`` requests depending on parameters.

        :param synths_only: Flag for freeing only child synths, or all children.
        """
        self.context.free_group_children(self, synths_only=synths_only)

    def query_tree(
        self,
        include_controls: bool = True,
        sync: bool = True,
    ) -> Union[Awaitable[Optional[QueryTreeGroup]], Optional[QueryTreeGroup]]:
        """
        Query the group's node tree.

        Emit ``/g_queryTree`` requests.

        :param include_controls: Flag for including synth control values.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        return cast(Union["AsyncServer", "Server"], self.context).query_tree(
            group=self, include_controls=include_controls, sync=sync
        )

    @property
    def children(self) -> List[Node]:
        """
        Get the group's children, as currently cached on the context.
        """
        children: List[Node] = []
        for id_ in cast("BaseServer", self.context)._node_children.get(self.id_, []):
            if id_ in cast("BaseServer", self.context)._node_children:
                children.append(Group(context=self.context, id_=id_))
            else:
                # cannot get synthdef name without running /g_queryTree
                children.append(Synth(context=self.context, id_=id_, synthdef=default))
        return children


@dataclasses.dataclass(frozen=True)
class RootNode(Group):
    r"""
    A root node.

    :param context: The root node's context.
    """

    def __post_init__(self) -> None:
        """
        Force ``id_`` to ``0``, and ``parallel`` to ``False``.
        """
        object.__setattr__(self, "id_", 0)
        object.__setattr__(self, "parallel", False)

    @property
    def _valid_add_actions(self) -> Container[int]:
        return (AddAction.ADD_TO_HEAD, AddAction.ADD_TO_TAIL)


@dataclasses.dataclass(frozen=True)
class Synth(Node):
    r"""
    A synth node.

    :param context: The synth's context.
    :param id\_: The synth's context ID.
    :param synthdef: The synth's SynthDef.
    """

    synthdef: SynthDef

    def get(self, *controls: Union[int, str], sync: bool = True) -> Union[
        Awaitable[Optional[Dict[Union[int, str], float]]],
        Optional[Dict[Union[int, str], float]],
    ]:
        """
        Get a control.

        Emit ``/s_get`` requests.

        :param controls: The control to get.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        return cast(Union["AsyncServer", "Server"], self.context).get_synth_controls(
            self, *controls, sync=sync
        )

    def get_range(
        self, control: Union[int, str], count: int, sync: bool = True
    ) -> Union[
        Awaitable[Optional[Sequence[Union[float, str]]]],
        Optional[Sequence[Union[float, str]]],
    ]:
        """
        Get a range of controls.

        Emit ``/s_get`` requests.

        :param control: The control to start reading at.
        :param count: The number of contiguous controls to get.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        return cast(
            Union["AsyncServer", "Server"], self.context
        ).get_synth_control_range(self, control, count, sync=sync)

    @property
    def _valid_add_actions(self) -> Container[int]:
        return (AddAction.ADD_AFTER, AddAction.ADD_BEFORE, AddAction.REPLACE)



import uuid
from unittest.mock import Mock, call

import pytest

from supriya.contexts import BusGroup, Server
from supriya.enums import CalculationRate
from supriya.patterns.events import BusAllocateEvent, BusFreeEvent, Priority

id_ = uuid.uuid4()


@pytest.mark.parametrize(
    "event, offset, expected",
    [(BusFreeEvent(id_), 0.0, [(0.0, Priority.START, BusFreeEvent(id_))])],
)
def test_expand(event, offset, expected):
    actual = event.expand(offset)
    assert actual == expected


def test_perform():
    context = Server().boot()
    spy = Mock(wraps=context)
    proxy_mapping = {}
    notes_mapping = {}
    # Allocate
    allocate_event = BusAllocateEvent(id_, calculation_rate="audio", channel_count=8)
    with context.at():
        allocate_event.perform(
            spy,
            proxy_mapping,
            current_offset=0.0,
            notes_mapping=notes_mapping,
            priority=Priority.START,
        )
    # Free
    free_event = BusFreeEvent(id_)
    with context.at():
        free_event.perform(
            spy,
            proxy_mapping,
            current_offset=0.0,
            notes_mapping=notes_mapping,
            priority=Priority.START,
        )
    assert proxy_mapping == {}
    assert notes_mapping == {}
    assert spy.mock_calls == [
        call.add_bus_group(calculation_rate=CalculationRate.AUDIO, count=8),
        call.free_bus_group(
            BusGroup(
                context=context,
                id_=16,
                calculation_rate=CalculationRate.AUDIO,
                count=8,
            )
        ),
    ]



"""
Tools for sending, receiving and handling OSC messages.
"""

from .asynchronous import AsyncOscProtocol
from .messages import OscBundle, OscMessage
from .protocols import (
    Capture,
    CaptureEntry,
    HealthCheck,
    OscCallback,
    OscProtocol,
    OscProtocolAlreadyConnected,
    OscProtocolOffline,
    find_free_port,
)
from .threaded import ThreadedOscProtocol

__all__ = [
    "AsyncOscProtocol",
    "Capture",
    "CaptureEntry",
    "HealthCheck",
    "OscBundle",
    "OscCallback",
    "OscMessage",
    "OscProtocol",
    "OscProtocolAlreadyConnected",
    "OscProtocolOffline",
    "ThreadedOscProtocol",
    "find_free_port",
]



# flake8: noqa
import os
import platform

import pytest

from supriya.ugens import (
    Demand,
    Dseries,
    Impulse,
    Out,
    SinOsc,
    SuperColliderSynthDef,
    SynthDefBuilder,
)


@pytest.fixture
def py_synthdef():
    with SynthDefBuilder() as builder:
        series = Dseries.dr(length=float("inf"), start=0, step=1)
        trigger = Impulse.kr(frequency=10)
        frequency = Demand.kr(trigger=trigger, source=series)
        Out.ar(bus=0, source=SinOsc.ar(frequency=frequency))
    py_synthdef = builder.build("foo")
    return py_synthdef


def test_demand_supriya_vs_bytes(py_synthdef):
    # fmt: off
    test_compiled_synthdef = bytes(
         b'SCgf'
         b'\x00\x00\x00\x02'
         b'\x00\x01'
             b'\x03foo'
                 b'\x00\x00\x00\x04'
                     b'\x7f\x80\x00\x00'
                     b'\x00\x00\x00\x00'
                     b'?\x80\x00\x00'
                     b'A \x00\x00'
                 b'\x00\x00\x00\x00'
                 b'\x00\x00\x00\x00'
                 b'\x00\x00\x00\x05'
                     b'\x07Dseries'
                        b'\x03'
                        b'\x00\x00\x00\x03'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x01'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x02'
                            b'\x03'
                     b'\x07Impulse'
                        b'\x01'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x03'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x01'
                            b'\x01'
                     b'\x06Demand'
                        b'\x01'
                        b'\x00\x00\x00\x03'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x01'
                            b'\x00\x00\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x01'
                            b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x00'
                            b'\x01'
                     b'\x06SinOsc'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x02'
                            b'\x00\x00\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x01'
                            b'\x02'
                     b'\x03Out'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x01'
                            b'\x00\x00\x00\x03'
                            b'\x00\x00\x00\x00'
                 b'\x00\x00'
    )
    # fmt: on
    py_compiled_synthdef = py_synthdef.compile()
    assert py_compiled_synthdef == test_compiled_synthdef


@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_demand_supriya_vs_sclang(py_synthdef):
    sc_synthdef = SuperColliderSynthDef(
        "foo",
        """
        var series, trigger, frequency;
        series = Dseries(0, 1, inf);
        trigger = Impulse.kr(frequency=10);
        frequency = Demand.kr(trigger, 0, series);
        Out.ar(0, SinOsc.ar(frequency));
        """,
    )
    sc_compiled_synthdef = sc_synthdef.compile()
    py_compiled_synthdef = py_synthdef.compile()
    assert py_compiled_synthdef == sc_compiled_synthdef



from .core import UGen, param, ugen


@ugen(ar=True, is_pure=True)
class BAllPass(UGen):
    """
    An all-pass filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> ball_pass = supriya.ugens.BAllPass.ar(
        ...     frequency=1200,
        ...     reciprocal_of_q=1,
        ...     source=source,
        ... )
        >>> ball_pass
        <BAllPass.ar()[0]>
    """

    source = param()
    frequency = param(1200.0)
    reciprocal_of_q = param(1.0)


@ugen(ar=True, is_pure=True)
class BBandPass(UGen):
    """
    A band-pass filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> bband_pass = supriya.ugens.BBandPass.ar(
        ...     bandwidth=1,
        ...     frequency=1200,
        ...     source=source,
        ... )
        >>> bband_pass
        <BBandPass.ar()[0]>
    """

    source = param()
    frequency = param(1200.0)
    bandwidth = param(1.0)


@ugen(ar=True, is_pure=True)
class BBandStop(UGen):
    """
    A band-stop filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> bband_stop = supriya.ugens.BBandStop.ar(
        ...     bandwidth=1,
        ...     frequency=1200,
        ...     source=source,
        ... )
        >>> bband_stop
        <BBandStop.ar()[0]>
    """

    source = param()
    frequency = param(1200.0)
    bandwidth = param(1.0)


@ugen(ar=True, is_pure=True)
class BHiCut(UGen):
    """
    A high-cut filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> bhi_cut = supriya.ugens.BHiCut.ar(
        ...     frequency=1200,
        ...     max_order=5,
        ...     order=2,
        ...     source=source,
        ... )
        >>> bhi_cut
        <BHiCut.ar()[0]>
    """

    source = param()
    frequency = param(1200.0)
    order = param(2.0)
    max_order = param(5.0)


@ugen(ar=True, is_pure=True)
class BHiPass(UGen):
    """
    A high-pass filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> bhi_pass = supriya.ugens.BHiPass.ar(
        ...     frequency=1200,
        ...     reciprocal_of_q=1,
        ...     source=source,
        ... )
        >>> bhi_pass
        <BHiPass.ar()[0]>
    """

    source = param()
    frequency = param(1200.0)
    reciprocal_of_q = param(1.0)


@ugen(ar=True, is_pure=True)
class BHiShelf(UGen):
    """
    A high-shelf filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> bhi_shelf = supriya.ugens.BHiShelf.ar(
        ...     gain=0,
        ...     frequency=1200,
        ...     reciprocal_of_s=1,
        ...     source=source,
        ... )
        >>> bhi_shelf
        <BHiShelf.ar()[0]>
    """

    source = param()
    frequency = param(1200.0)
    reciprocal_of_s = param(1.0)
    gain = param(0.0)


@ugen(ar=True, is_pure=True)
class BLowCut(UGen):
    """
    A low-cut filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> blow_cut = supriya.ugens.BLowCut.ar(
        ...     frequency=1200,
        ...     max_order=5,
        ...     order=2,
        ...     source=source,
        ... )
        >>> blow_cut
        <BLowCut.ar()[0]>
    """

    source = param()
    frequency = param(1200.0)
    order = param(2.0)
    max_order = param(5.0)


@ugen(ar=True, is_pure=True)
class BLowPass(UGen):
    """
    A low-pass filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> blow_pass = supriya.ugens.BLowPass.ar(
        ...     frequency=1200,
        ...     reciprocal_of_q=1,
        ...     source=source,
        ... )
        >>> blow_pass
        <BLowPass.ar()[0]>
    """

    source = param()
    frequency = param(1200.0)
    reciprocal_of_q = param(1.0)


@ugen(ar=True, is_pure=True)
class BLowShelf(UGen):
    """
    A low-shelf filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> blow_shelf = supriya.ugens.BLowShelf.ar(
        ...     frequency=1200,
        ...     gain=0,
        ...     reciprocal_of_s=1,
        ...     source=source,
        ... )
        >>> blow_shelf
        <BLowShelf.ar()[0]>
    """

    source = param()
    frequency = param(1200.0)
    reciprocal_of_s = param(1.0)
    gain = param(0.0)


@ugen(ar=True, is_pure=True)
class BPeakEQ(UGen):
    """
    A parametric equalizer.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> bpeak_eq = supriya.ugens.BPeakEQ.ar(
        ...     frequency=1200,
        ...     gain=0,
        ...     reciprocal_of_q=1,
        ...     source=source,
        ... )
        >>> bpeak_eq
        <BPeakEQ.ar()[0]>
    """

    source = param()
    frequency = param(1200.0)
    reciprocal_of_q = param(1.0)
    gain = param(0.0)
